use n::packet::*;
use n::ip;

#[inline(always)]
pub fn extract_vhl(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(1);
	packet.context.set_field("ipv4::vhl", field);
	packet
}

#[inline(always)]
pub fn extract_dscp_ecn(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(1);
	packet.context.set_field("ipv4::dscp_ecn", field);
	packet
}

#[inline(always)]
pub fn extract_identification(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(2);
	packet.context.set_field("ipv4::identification", field);
	packet
}

#[inline(always)]
pub fn extract_fragment(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(2);
	packet.context.set_field("ipv4::fragment", field);
	packet
}

#[inline(always)]
pub fn extract_header_checksum(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(2);
	packet.context.set_field("ipv4::header_checksum", field);
	packet
}

#[inline(always)]
pub fn extract_source(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(4);
	packet.context.set_field("ipv4::source", field);
	packet
}

#[inline(always)]
pub fn extract_destination(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(4);
	packet.context.set_field("ipv4::destination", field);
	packet
}

#[inline(always)]
pub fn extract_version_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let ihl;
	{
		let field = packet.context.get_field("ipv4::vhl");
		ihl = field.read_u8(0);
	}
	packet.context.set_u8("ipv4::version_u8", (ihl >> 4) & 0x0F);
	packet
}

#[inline(always)]
pub fn extract_total_length(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::total_length", "ipv4::total_length");
	ip::extract_total_length(packet)
}

#[inline(always)]
pub fn extract_total_length_u16(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::total_length_u16", "ipv4::total_length_u16");
	ip::extract_total_length_u16(packet)
}

#[inline(always)]
pub fn extract_hop_limit(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::hop_limit", "ipv4::hop_limit");
	ip::extract_hop_limit(packet)
}

#[inline(always)]
pub fn extract_hop_limit_u8(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::hop_limit_u8", "ipv4::hop_limit_u8");
	ip::extract_hop_limit_u8(packet)
}

#[inline(always)]
pub fn extract_protocol(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::protocol", "ipv4::protocol");
	ip::extract_protocol(packet)
}

#[inline(always)]
pub fn extract_protocol_u8(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::protocol_u8", "ipv4::protocol_u8");
	ip::extract_protocol_u8(packet)
}

#[inline(always)]
pub fn extract_header_length_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let ihl;
	{
		let field = packet.context.get_field("ipv4::vhl");
		ihl = field.read_u8(0);
	}
	packet.context.set_u8("ipv4::header_length_u8", (ihl & 0x0F) << 2);
	packet
}

#[inline(always)]
pub fn extract_dscp_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let dscp_ecn;
	{
		let field = packet.context.get_field("ipv4::dscp_ecn");
		dscp_ecn = field.read_u8(0);
	}
	packet.context.set_u8("ipv4::dscp_u8", (dscp_ecn >> 2) & 0x3F);
	packet.context.set_alias("ip::dscp_u8", "ipv4::dscp_u8");
	packet
}

#[inline(always)]
pub fn extract_ecn_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let dscp_ecn;
	{
		let field = packet.context.get_field("ipv4::dscp_ecn");
		dscp_ecn = field.read_u8(0);
	}
	packet.context.set_u8("ipv4::ecn_u8", dscp_ecn & 0x03);
	packet.context.set_alias("ip::ecn_u8", "ipv4::ecn_u8");
	packet
}

#[inline(always)]
pub fn extract_fragment_flag_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let fragment;
	{
		let field = packet.context.get_field("ipv4::fragment");
		fragment = field.read_u8(0);
	}
	packet.context.set_u8("ipv4::fragment_flag_u8", (fragment >> 5) & 0x03);
	packet
}

#[inline(always)]
pub fn extract_fragment_offset_u16(packet : Packet) -> Packet {
	let mut packet = packet;
	let fragment;
	{
		let field = packet.context.get_field("ipv4::fragment");
		fragment = field.read_u16(0);
	}
	packet.context.set_u16("ipv4::fragment_offset_u16", (fragment) & 0x1FFF);
	packet
}

#[inline(always)]
pub fn extract_header_checksum_u16(packet : Packet) -> Packet {
	let mut packet = packet;
	let checksum;
	{
		let field = packet.context.get_field("ipv4::header_checksum");
		checksum = field.read_u16(0);
	}
	packet.context.set_u16("ipv4::header_checksum_u16", checksum);
	packet
}

#[inline(always)]
pub fn extract_option(packet : Packet) -> Packet {	
	let ihl;
	{
		let field = packet.context.get_field("ipv4::vhl");
		ihl = field.read_u8(0) & 0x0F;
	}
	
	assert!(ihl >= 5);
	let option_bytes = ((ihl-5) << 2) as usize;
	let (mut packet,field) = packet.next_packet(option_bytes);
	packet.context.set_field("ipv4::option", field);
	packet
}

#[inline(always)]
pub fn verify_header_checksum(packet : Packet) -> Packet {
    let mut packet = packet;
    let header_length;
    {
        header_length = (packet.read_u8(0) & 0xFF) << 2;
    }
    let result = packet.ones_complement_sum_u16(header_length as usize);
    packet.context.set_bool("ipv4::correct_csum_bool", 
        (result == 0xFFFF) || (result == 0x0000));
    packet
}

#[inline(always)]
pub fn prepare_pseudo_header_u16(packet : Packet) -> Packet {
    let mut sum : u32;
    let playload_len;
    {
        let ipv4_src = packet.context.get_field("ipv4::source");
        let ipv4_dst = packet.context.get_field("ipv4::destination");
        let ipv4_proto = packet.context.get_field("ipv4::protocol").read_u8(0);
        
        let ipv4_hlen;
    	{
    		let field = packet.context.get_field("ipv4::vhl");
    		ipv4_hlen = field.read_u8(0);
    	}
    	let ipv4_hlen = (ipv4_hlen & 0x0F) << 2;
        let ipv4_tlen = packet.context.get_field("ipv4::total_length").read_u16(0);
        
        playload_len = (ipv4_tlen - (ipv4_hlen as u16)) as u32;
        sum = playload_len;
        sum = sum + (ipv4_proto as u16) as u32;
        sum = sum + ipv4_src.ones_complement_sum_u16(ipv4_src.len()) as u32;
        sum = sum + ipv4_dst.ones_complement_sum_u16(ipv4_dst.len()) as u32;
        
        while(sum & 0xFFFF0000) > 0 {
            sum = (sum >> 16) + (sum & 0x0000FFFF);
        }
    }
    let mut packet = packet;
    packet.context.set_u16("ipv4::pseudo_csum_u16", sum as u16);
    packet.context.set_alias("ip::pseudo_csum_u16", "ipv4::pseudo_csum_u16");
    packet.context.set_u32("ipv4::payload_length_u32", playload_len);
    packet.context.set_alias("ip::payload_length_u32", "ipv4::payload_length_u32");
    packet
}

#[inline(always)]
pub fn extract(packet : Packet) -> Packet {
    let packet = verify_header_checksum(packet);
	let packet = extract_vhl(packet);
	let packet = extract_dscp_ecn(packet);
	let packet = extract_total_length(packet);
	let packet = extract_identification(packet);
	let packet = extract_fragment(packet);
	let packet = extract_hop_limit(packet);
	let packet = extract_protocol(packet);
	let packet = extract_header_checksum(packet);
	let packet = extract_source(packet);
	let packet = extract_destination(packet);
	let packet = extract_option(packet);
	let packet = prepare_pseudo_header_u16(packet);
	
	packet
}