use n::packet::*;
use n::ip;

#[inline(always)]
pub fn extract_vtf(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(4);
	packet.context.set_field("ipv6::vtf", field);
	packet
}

#[inline(always)]
pub fn extract_version_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let vtf;
	{
		let field = packet.context.get_field("ipv6::vtf");
		vtf = field.read_u8(0);
	}
	packet.context.set_u8("ipv6::version_u8", (vtf >> 4) & 0x0F);
	packet
}

#[inline(always)]
pub fn extract_dscp_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let vtf;
	{
		let field = packet.context.get_field("ipv6::vtf");
		vtf = field.read_u16(0);
	}
	let dscp_ecn = ((vtf >> 4) & 0xFF) as u8;
	packet.context.set_u8("ipv6::dscp_u8", (dscp_ecn >> 2) & 0x3F);
	packet.context.set_alias("ip::dscp_u8", "ipv6::dscp_u8");
	packet
}

#[inline(always)]
pub fn extract_ecn_u8(packet : Packet) -> Packet {
	let mut packet = packet;
	let vtf;
	{
		let field = packet.context.get_field("ipv6::vtf");
		vtf = field.read_u16(0);
	}
	let dscp_ecn = ((vtf >> 4) & 0xFF) as u8;
	packet.context.set_u8("ipv6::ecn_u8", dscp_ecn & 0x03);
	packet.context.set_alias("ip::ecn_u8", "ipv6::ecn_u8");
	packet
}

#[inline(always)]
pub fn extract_flow_u32(packet : Packet) -> Packet {
	let mut packet = packet;
	let vtf;
	{
		let field = packet.context.get_field("ipv6::vtf");
		vtf = field.read_u32(0);
	}
	packet.context.set_u32("ipv6::flow_u32", vtf & 0x000FFFFF);
	packet
}

#[inline(always)]
pub fn extract_total_length(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::total_length", "ipv6::total_length");
	ip::extract_total_length(packet)
}

#[inline(always)]
pub fn extract_total_length_u16(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::total_length_u16", "ipv6::total_length_u16");
	ip::extract_total_length_u16(packet)
}

#[inline(always)]
pub fn extract_protocol(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::protocol", "ipv6::protocol");
	ip::extract_protocol(packet)
}

#[inline(always)]
pub fn extract_protocol_u8(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::protocol_u8", "ipv6::protocol_u8");
	ip::extract_protocol_u8(packet)
}

#[inline(always)]
pub fn extract_hop_limit(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::hop_limit", "ipv6::hop_limit");
	ip::extract_hop_limit(packet)
}

#[inline(always)]
pub fn extract_hop_limit_u8(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("ip::hop_limit_u8", "ipv6::hop_limit_u8");
	ip::extract_hop_limit_u8(packet)
}

#[inline(always)]
pub fn extract_source(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(16);
	packet.context.set_field("ipv6::source", field);
	packet
}

#[inline(always)]
pub fn extract_destination(packet : Packet) -> Packet {
	let (mut packet,field) = packet.next_packet(16);
	packet.context.set_field("ipv6::destination", field);
	packet
}

#[inline(always)]
pub fn prepare_pseudo_header_u16(packet : Packet) -> Packet {
    let mut sum : u32;
    let ipv6_body;
    {
        let ipv6_src = packet.context.get_field("ipv6::source");
        let ipv6_dst = packet.context.get_field("ipv6::destination");
        let ipv6_proto = packet.context.get_field("ipv6::protocol").read_u8(0);
        let ipv6_hlen = packet.context.get_u32("ipv6::header_length_u32");
        let ipv6_tlen = packet.context.get_u32("ipv6::total_length_u32");
        
        ipv6_body = ipv6_tlen - ipv6_hlen;
        sum = (ipv6_body & 0x0000FFFF) + (ipv6_body >> 16);
        sum = sum + ipv6_proto as u32;
        sum = sum + ipv6_src.ones_complement_sum_u16(ipv6_src.len()) as u32;
        sum = sum + ipv6_dst.ones_complement_sum_u16(ipv6_dst.len()) as u32;
        
        while(sum & 0xFFFF0000) > 0 {
            sum = (sum >> 16) + (sum & 0x0000FFFF);
        }
    }
    let mut packet = packet;
    packet.context.set_u16("ipv6::pseudo_csum_u16", sum as u16);
    packet.context.set_alias("ip::pseudo_csum_u16", "ipv6::pseudo_csum_u16");
    packet.context.set_u32("ipv6::payload_length_u32", ipv6_body);
    packet.context.set_alias("ip::payload_length_u32", "ipv6::payload_length_u32");
    packet
}

#[inline(always)]
pub fn has_more_options(packet : Packet) -> Packet {
    
    let has_next = match packet.context.get_field("ip::protocol").read_u8(0) {
        0x00 | //Hop-by-Hop Options
        0x2B | //Routing (Type 0)
        0x2C | //Fragment
        0x3C | //Destination Options
        0x33 | //Authentication
        0x32 //Encapsulating Security Payload
        => true,
        _ => false,
    };
    let mut packet = packet;
    packet.context.set_bool_released("ipv6::more_options_bool", has_next);
    
	packet
}

#[inline(always)]
pub fn prepare_option_parsing(packet : Packet) -> Packet {
    let mut packet = packet;
    let ipv6_hlen = 
        match packet.context.get_u32_released("ipv6::header_length_u32") {
            Some(x) => x,
            _ => 0, //ipv6 hlen does not include default header
        };
    let ipv6_tlen = 
        match packet.context.get_u32_released("ipv6::total_length_u32") {
            Some(x) => x,
            _ => packet.context.get_field("ipv6::total_length").read_u16(0) as u32
        };
    packet.context.set_u32_released("ipv6::header_length_u32", ipv6_hlen);
    packet.context.set_u32_released("ipv6::total_length_u32", ipv6_tlen);
    
	packet
}

#[inline(always)]
pub fn extract_current_option(packet : Packet) -> Packet {
    let ipv6_hlen = packet.context.get_u32("ipv6::header_length_u32");
    
    let (packet, next_header) = packet.next_packet(1);
    let (packet, ext_len) = packet.next_packet(1);
    let ext_len = ext_len.read_u8(0) << 3; //8 octets
    let mut packet = packet.strip(6);
    packet.context.set_field_released("ipv6::protocol",next_header);
    packet.context.set_u32_released("ipv6::header_length_u32", ipv6_hlen + 8 + ext_len as u32);
    packet.strip(ext_len as usize)
}

#[inline(always)]
pub fn extract(packet : Packet) -> Packet {
	let packet = extract_vtf(packet);
	let packet = extract_total_length(packet);
	let packet = extract_protocol(packet);
	let packet = extract_hop_limit(packet);
	let packet = extract_source(packet);
	let packet = extract_destination(packet);
	let packet = prepare_option_parsing(packet);
	
	let mut packet = packet;
	loop {
	    packet = has_more_options(packet);
	    if !packet.context.get_bool("ipv6::more_options_bool") {
	        break;
	    }
	    packet = extract_current_option(packet);
	}
	
	packet = prepare_pseudo_header_u16(packet);
	
	packet
}