use n::packet::*;
use n::l4;
use std::cmp::min;

#[inline(always)]
pub fn extract_source(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("l4::source", "tcp::source");
	l4::extract_source(packet)
}

#[inline(always)]
pub fn extract_source_u16(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("l4::source_u16", "tcp::source_u16");
	l4::extract_source_u16(packet)
}

#[inline(always)]
pub fn extract_destination(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("l4::destination", "tcp::destination");
	l4::extract_destination(packet)
}

#[inline(always)]
pub fn extract_destination_u16(packet : Packet) -> Packet {
    let mut packet = packet;
    packet.context.set_alias("l4::destination_u16", "tcp::destination_u16");
	l4::extract_destination_u16(packet)
}

#[inline(always)]
pub fn extract_sequence(packet : Packet) -> Packet {
    let (mut packet,field) = packet.next_packet(4);
	packet.context.set_field("tcp::sequence", field);
	packet
}

#[inline(always)]
pub fn extract_sequence_u32(packet : Packet) -> Packet {
    let mut packet = packet;
	let seq;
	{
		let field = packet.context.get_field("tcp::sequence");
		seq = field.read_u32(0);
	}
	packet.context.set_u32("tcp::sequence_u32", seq);
	packet
}

#[inline(always)]
pub fn extract_acknowledgment(packet : Packet) -> Packet {
    let (mut packet,field) = packet.next_packet(4);
	packet.context.set_field("tcp::acknowledgment", field);
	packet
}

#[inline(always)]
pub fn extract_acknowledgment_u32(packet : Packet) -> Packet {
    let mut packet = packet;
	let ack;
	{
		let field = packet.context.get_field("tcp::acknowledgment");
		ack = field.read_u32(0);
	}
	packet.context.set_u32("tcp::acknowledgment_u32", ack);
	packet
}

#[inline(always)]
pub fn extract_offset_flags(packet : Packet) -> Packet {
    let (mut packet,field) = packet.next_packet(2);
	packet.context.set_field("tcp::offset_flags", field);
	packet
}

#[inline(always)]
pub fn extract_header_length_u8(packet : Packet) -> Packet {
    let mut packet = packet;
	let offset;
	{
		let field = packet.context.get_field("tcp::offset_flags");
		offset = field.read_u8(0) >> 4;
	}
	packet.context.set_u8("tcp::header_length_u8", offset << 2);
	packet
}

#[inline(always)]
pub fn extract_flags_u16(packet : Packet) -> Packet {
    let mut packet = packet;
	let flags;
	{
		let field = packet.context.get_field("tcp::offset_flags");
		flags = field.read_u16(0) & 0x01FF;
	}
	packet.context.set_u16("tcp::flags_u16", flags);
	packet
}

#[inline(always)]
pub fn extract_window(packet : Packet) -> Packet {
    let (mut packet,field) = packet.next_packet(2);
	packet.context.set_field("tcp::window", field);
	packet
}

#[inline(always)]
pub fn extract_window_u16(packet : Packet) -> Packet {
    let mut packet = packet;
	let window;
	{
		let field = packet.context.get_field("tcp::window");
		window = field.read_u16(0);
	}
	packet.context.set_u16("tcp::window_u16", window);
	packet
}

#[inline(always)]
pub fn extract_checksum(packet : Packet) -> Packet {
    let (mut packet,field) = packet.next_packet(2);
	packet.context.set_field("tcp::checksum", field);
	packet
}

#[inline(always)]
pub fn extract_checksum_u16(packet : Packet) -> Packet {
    let mut packet = packet;
	let checksum;
	{
		let field = packet.context.get_field("tcp::checksum");
		checksum = field.read_u16(0);
	}
	packet.context.set_u16("tcp::checksum_u16", checksum);
	packet
}

#[inline(always)]
pub fn extract_urgent(packet : Packet) -> Packet {
    let (mut packet,field) = packet.next_packet(2);
	packet.context.set_field("tcp::urgent", field);
	packet
}

#[inline(always)]
pub fn extract_urgent_u16(packet : Packet) -> Packet {
    let mut packet = packet;
	let urgent;
	{
		let field = packet.context.get_field("tcp::urgent");
		urgent = field.read_u16(0);
	}
	packet.context.set_u16("tcp::urgent_u16", urgent);
	packet
}

#[inline(always)]
pub fn prepare_option_parsing(packet : Packet) -> Packet {
    let mut packet = packet;
    let tcp_hlen = packet.context.get_field("tcp::offset_flags").read_u8(0);
    let tcp_hlen = (tcp_hlen >> 4) << 2;
    assert!(tcp_hlen >= 20);
    packet.context.set_u8("tcp::remaining_options_u8", tcp_hlen - 20);
    
	packet
}

#[inline(always)]
pub fn has_more_options(packet : Packet) -> Packet {
    let mut packet = packet;
    let remaining = packet.context.get_u8("tcp::remaining_options_u8");
    packet.context.set_bool_released("tcp::more_options_bool", remaining > 0);
    
	packet
}

#[inline(always)]
pub fn extract_current_option(packet : Packet) -> Packet {
    let mut remaining = packet.context.get_u8("tcp::remaining_options_u8");
    assert!(remaining > 0);
    let (mut packet, kind) = packet.next_packet(1);
    remaining = remaining - 1;
    match kind.read_u8(0) {
        0 => { 
            packet.context.set_u8_released("tcp::remaining_options_u8", 0); 
            packet.strip(remaining as usize)
            }, //end of list
        1 => { 
            packet.context.set_u8_released("tcp::remaining_options_u8", remaining);
            packet
            }, //no-op
        _ => {
            let (mut packet, size_field) = packet.next_packet(1);
            remaining = remaining - 1;
            let size = min(size_field.read_u8(0) - 2, remaining); //kind and size itself
            remaining = remaining - size;
            
            packet.context.set_u8_released("tcp::remaining_options_u8", remaining); 
            packet.strip(size as usize)
            }, //end of list
    }
}

#[inline(always)]
pub fn verify_header_checksum(packet : Packet) -> Packet {
    let mut packet = packet;
    let payload_length = min(
        packet.len(),
        packet.context.get_u32("ip::payload_length_u32") as usize
    );
    let mut sum = packet.ones_complement_sum_u16(payload_length) as u32;
    sum = sum + packet.context.get_u16("ip::pseudo_csum_u16") as u32;
    sum = (sum >> 16) + (sum & 0x0000FFFF);
    let result = sum as u16;
    packet.context.set_u16("tcp::current_csum_u16", result);
    packet.context.set_bool("tcp::correct_csum_bool", 
        (result == 0xFFFF) || (result == 0x0000));
    packet
}

#[inline(always)]
pub fn extract(packet : Packet) -> Packet {
    let packet = verify_header_checksum(packet);
    let packet = extract_source(packet);
	let packet = extract_destination(packet);
	let packet = extract_sequence(packet);
	let packet = extract_acknowledgment(packet);
	let packet = extract_offset_flags(packet);
	let packet = extract_window(packet);
	let packet = extract_checksum(packet);
	let packet = extract_urgent(packet);
	
	let packet = prepare_option_parsing(packet);
	
	let mut packet = packet;
	loop {
	    packet = has_more_options(packet);
	    if !packet.context.get_bool("tcp::more_options_bool") {
	        break;
	    }
	    packet = extract_current_option(packet);
	}
	
	packet
}